import logging
import multiprocessing as mp

import rethinkdb as r

from bigchaindb import Bigchain
from bigchaindb.voter import Voter
from bigchaindb.block import Block


logger = logging.getLogger(__name__)


class Processes(object):

    def __init__(self):
        # initialize the class
        self.q_new_block = mp.Queue()
        self.q_new_transaction = mp.Queue()

    def map_backlog(self):
        # listen to changes on the backlog and redirect the changes
        # to the correct queues

        # create a bigchain instance
        b = Bigchain()

        for change in r.table('backlog').changes().run(b.conn):

            # insert
            if change['old_val'] is None:
                self.q_new_transaction.put(change['new_val'])

            # delete
            if change['new_val'] is None:
                pass

            # update
            if change['new_val'] is not None and change['old_val'] is not None:
                pass

    def map_bigchain(self):
        # listen to changes on the bigchain and redirect the changes
        # to the correct queues

        # create a bigchain instance
        b = Bigchain()

        for change in r.table('bigchain').changes().run(b.conn):

            # insert
            if change['old_val'] is None:
                self.q_new_block.put(change['new_val'])

            # delete
            elif change['new_val'] is None:
                pass

            # update
            elif change['new_val'] is not None and change['old_val'] is not None:
                pass

    def start(self):
        # instantiate block and voter
        block = Block(self.q_new_transaction)

        # initialize the processes
        p_map_bigchain = mp.Process(name='bigchain_mapper', target=self.map_bigchain)
        p_map_backlog = mp.Process(name='backlog_mapper', target=self.map_backlog)
        p_block = mp.Process(name='block', target=block.start)
        p_voter = Voter(self.q_new_block)

        # start the processes
        logger.info('starting bigchain mapper')
        p_map_bigchain.start()
        logger.info('starting backlog mapper')
        p_map_backlog.start()
        logger.info('starting block')
        p_block.start()

        logger.info('starting voter')
        p_voter.start()



"""Utils to configure Bigchain.

By calling `file_config`, the global configuration (stored in
`bigchain.config`) will be updated with the values contained in the
configuration file.

Note that there is a precedence in reading configuration values:
 - [not yet] command line;
 - local config file;
 - environment vars;
 - default config file (contained in `bigchain.__init__`).
"""

import os
import copy
import json
import logging
import collections

import bigchaindb

logger = logging.getLogger(__name__)
CONFIG_DEFAULT_PATH = os.environ.setdefault(
    'BIGCHAINDB_CONFIG_PATH',
    os.path.join(os.path.expanduser('~'), '.bigchaindb'),
)


# Thanks Alex <3
# http://stackoverflow.com/a/3233356/597097
def update(d, u):
    """Recursively update a mapping."""
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def file_config(filename=None):
    """Read a configuration file and merge it with the default configuration.

    Args:
        filename (str): the JSON file with the configuration. Defaults to ``None``.
            If ``None``, the HOME of the current user and the string ``.bigchaindb`` will be used.

    Note:
        The function merges the values in ``filename`` with the **default configuration**,
        so any update made to ``bigchaindb.config`` will be lost.
    """
    if not filename:
        filename = CONFIG_DEFAULT_PATH

    with open(filename) as f:
        newconfig = json.load(f)

    dict_config(newconfig)
    logger.info('Configuration loaded from `{}`'.format(filename))


def dict_config(newconfig):
    """Merge the provided configuration with the default one.

    Args:
        newconfig (dict): a dictionary with the configuration to load.

    Note:
        The function merges ``newconfig`` with the **default configuration**, so any
        update made to ``bigchaindb.config`` will be lost.
    """
    bigchaindb.config = copy.deepcopy(bigchaindb._config)
    update(bigchaindb.config, newconfig)
    bigchaindb.config['CONFIGURED'] = True


def write_config(newconfig, filename=None):
    """Write the provided configuration to a specific location.

    Args:
        newconfig (dict): a dictionary with the configuration to load.
        filename (str): the name of the file that will store the new configuration. Defaults to ``None``.
            If ``None``, the HOME of the current user and the string ``.bigchaindb`` will be used.
    """
    if not filename:
        filename = CONFIG_DEFAULT_PATH

    with open(filename, 'w') as f:
        json.dump(newconfig, f)


def autoconfigure():
    """Run ``file_config`` if the module has not been initialized.
    """
    if bigchaindb.config.get('CONFIGURED'):
        return
    try:
        file_config()
    except FileNotFoundError:
        logger.warning('Cannot find your config file. Run `bigchaindb configure` to create one')




import rethinkdb as r
import time
import random
import json
import rapidjson

from datetime import datetime

import bigchaindb
from bigchaindb import config_utils
from bigchaindb import exceptions
from bigchaindb.crypto import hash_data, PublicKey, PrivateKey, generate_key_pair


class GenesisBlockAlreadyExistsError(Exception):
    pass


class KeypairNotFoundException(Exception):
    pass


class Bigchain(object):
    """Bigchain API

    Create, read, sign, write transactions to the database
    """

    def __init__(self, host=None, port=None, dbname=None,
                 public_key=None, private_key=None, keyring=[]):
        """Initialize the Bigchain instance

        There are three ways in which the Bigchain instance can get its parameters.
        The order by which the parameters are chosen are:

            1. Setting them by passing them to the `__init__` method itself.
            2. Setting them as environment variables
            3. Reading them from the `config.json` file.

        Args:
            host (str): hostname where the rethinkdb is running.
            port (int): port in which rethinkb is running (usually 28015).
            dbname (str): the name of the database to connect to (usually bigchain).
            public_key (str): the base58 encoded public key for the ECDSA secp256k1 curve.
            private_key (str): the base58 encoded private key for the ECDSA secp256k1 curve.
            keyring (list[str]): list of base58 encoded public keys of the federation nodes.

        """
        config_utils.autoconfigure()
        self.host = host or bigchaindb.config['database']['host']
        self.port = port or bigchaindb.config['database']['port']
        self.dbname = dbname or bigchaindb.config['database']['name']
        self.me = public_key or bigchaindb.config['keypair']['public']
        self.me_private = private_key or bigchaindb.config['keypair']['private']
        self.federation_nodes = keyring or bigchaindb.config['keyring']

        if not self.me or not self.me_private:
            raise KeypairNotFoundException()

        self._conn = None

    @property
    def conn(self):
        if not self._conn:
            self._conn = self.reconnect()
        return self._conn

    def reconnect(self):
        return r.connect(host=self.host, port=self.port, db=self.dbname)

    def create_transaction(self, current_owner, new_owner, tx_input, operation, payload=None):
        """Create a new transaction

        A transaction in the bigchain is a transfer of a digital asset between two entities represented
        by public keys.

        Currently the bigchain supports two types of operations:

            `CREATE` - Only federation nodes are allowed to use this operation. In a create operation
            a federation node creates a digital asset in the bigchain and assigns that asset to a public
            key. The owner of the private key can then decided to transfer this digital asset by using the
            `transaction id` of the transaction as an input in a `TRANSFER` transaction.

            `TRANSFER` - A transfer operation allows for a transfer of the digital assets between entities.

        Args:
            current_owner (str): base58 encoded public key of the current owner of the asset.
            new_owner (str): base58 encoded public key of the new owner of the digital asset.
            tx_input (str): id of the transaction to use as input.
            operation (str): Either `CREATE` or `TRANSFER` operation.
            payload (Optional[dict]): dictionary with information about asset.

        Returns:
            dict: unsigned transaction.


        Raises:
            TypeError: if the optional ``payload`` argument is not a ``dict``.
        """
        data = None
        if payload is not None:
            if isinstance(payload, dict):
                hash_payload = hash_data(self.serialize(payload))
                data = {
                    'hash': hash_payload,
                    'payload': payload
                }
            else:
                raise TypeError('`payload` must be an dict instance')

        hash_payload = hash_data(self.serialize(payload))
        data = {
            'hash': hash_payload,
            'payload': payload
        }

        tx = {
            'current_owner': current_owner,
            'new_owner': new_owner,
            'input': tx_input,
            'operation': operation,
            'timestamp': self.timestamp(),
            'data': data
        }

        # serialize and convert to bytes
        tx_serialized = self.serialize(tx)
        tx_hash = hash_data(tx_serialized)

        # create the transaction
        transaction = {
            'id': tx_hash,
            'transaction': tx
        }

        return transaction

    def sign_transaction(self, transaction, private_key):
        """Sign a transaction

        A transaction signed with the `current_owner` corresponding private key.

        Args:
            transaction (dict): transaction to sign.
            private_key (str): base58 encoded private key to create a signature of the transaction.

        Returns:
            dict: transaction with the `signature` field included.

        """
        private_key = PrivateKey(private_key)
        signature = private_key.sign(self.serialize(transaction))
        signed_transaction = transaction.copy()
        signed_transaction.update({'signature': signature})
        return signed_transaction

    def verify_signature(self, signed_transaction):
        """Verify the signature of a transaction

        A valid transaction should have been signed `current_owner` corresponding private key.

        Args:
            signed_transaction (dict): a transaction with the `signature` included.

        Returns:
            bool: True if the signature is correct, False otherwise.

        """
        data = signed_transaction.copy()

        # if assignee field in the transaction, remove it
        if 'assignee' in data:
            data.pop('assignee')

        signature = data.pop('signature')
        public_key_base58 = signed_transaction['transaction']['current_owner']
        public_key = PublicKey(public_key_base58)
        return public_key.verify(self.serialize(data), signature)

    def write_transaction(self, signed_transaction):
        """Write the transaction to bigchain.

        When first writing a transaction to the bigchain the transaction will be kept in a backlog until
        it has been validated by the nodes of the federation.

        Args:
            singed_transaction (dict): transaction with the `signature` included.

        Returns:
            dict: database response
        """

        # we will assign this transaction to `one` node. This way we make sure that there are no duplicate
        # transactions on the bigchain

        if self.federation_nodes:
            assignee = random.choice(self.federation_nodes)
        else:
            # I am the only node
            assignee = self.me

        # update the transaction
        signed_transaction.update({'assignee': assignee})

        # write to the backlog
        response = r.table('backlog').insert(signed_transaction, durability='soft').run(self.conn)
        return response

    # TODO: the same `txid` can be in two different blocks
    def get_transaction(self, txid):
        """Retrieve a transaction with `txid` from bigchain.

        Queries the bigchain for a transaction that was already included in a block.

        Args:
            txid (str): transaction id of the transaction to query

        Returns:
            A dict with the transaction details if the transaction was found.

            If no transaction with that `txid` was found it returns `None`

        """
        response = r.table('bigchain').concat_map(lambda doc: doc['block']['transactions'])\
            .filter(lambda transaction: transaction['id'] == txid).run(self.conn)

        # transaction ids should be unique
        transactions = list(response)
        if transactions:
            if len(transactions) != 1:
                raise Exception('Transaction ids should be unique. There is a problem with the chain')
            else:
                return transactions[0]
        else:
            return None

    def get_tx_by_payload_hash(self, payload_hash):
        """Retrieves transactions related to a digital asset.

        When creating a transaction one of the optional arguments is the `payload`. The payload is a generic
        dict that contains information about the digital asset.

        To make it easy to query the bigchain for that digital asset we create a sha3-256 hash of the
        serialized payload and store it with the transaction. This makes it easy for developers to keep track
        of their digital assets in bigchain.

        Args:
            payload_hash (str): sha3-256 hash of the serialized payload.

        Returns:
            A list of transactions containing that payload. If no transaction exists with that payload it
            returns `None`

        """
        cursor = r.table('bigchain')\
            .get_all(payload_hash, index='payload_hash')\
            .run(self.conn)

        transactions = list(cursor)
        return transactions

    def get_spent(self, txid):
        """Check if a `txid` was already used as an input.

        A transaction can be used as an input for another transaction. Bigchain needs to make sure that a
        given `txid` is only used once.

        Args:
            txid (str): transaction id.

        Returns:
            The transaction that used the `txid` as an input if it exists else it returns `None`

        """

        # checks if an input was already spent
        # checks if the bigchain has any transaction with input `transaction_id`
        response = r.table('bigchain').concat_map(lambda doc: doc['block']['transactions'])\
            .filter(lambda transaction: transaction['transaction']['input'] == txid).run(self.conn)

        # a transaction_id should have been spent at most one time
        transactions = list(response)
        if transactions:
            if len(transactions) != 1:
                raise Exception('`{}` was spent more then once. There is a problem with the chain'.format(
                    txid))
            else:
                return transactions[0]
        else:
            return None

    def get_owned_ids(self, owner):
        """Retrieve a list of `txids` that can we used has inputs.

        Args:
            owner (str): base58 encoded public key.

        Returns:
            list: list of `txids` currently owned by `owner`

        """
        response = r.table('bigchain')\
                    .concat_map(lambda doc: doc['block']['transactions'])\
                    .filter({'transaction': {'new_owner': owner}})\
                    .pluck('id')['id']\
                    .run(self.conn)
        owned = []

        # remove all inputs already spent
        for tx_input in list(response):
            if not self.get_spent(tx_input):
                owned.append(tx_input)

        return owned

    def validate_transaction(self, transaction):
        """Validate a transaction.

        Args:
            transaction (dict): transaction to validate.

        Returns:
            The transaction if the transaction is valid else it raises and exception
            describing the reason why the transaction is invalid.

        Raises:
            OperationError: if the transaction operation is not supported
            TransactionDoesNotExist: if the input of the transaction is not found
            TransactionOwnerError: if the new transaction is using an input it doesn't own
            DoubleSpend: if the transaction is a double spend
            InvalidHash: if the hash of the transaction is wrong
            InvalidSignature: if the signature of the transaction is wrong
        """
        # If the operation is CREATE the transaction should have no inputs and should be signed by a
        # federation node
        if transaction['transaction']['operation'] == 'CREATE':
            if transaction['transaction']['input']:
                raise ValueError('A CREATE operation has no inputs')
            if transaction['transaction']['current_owner'] not in self.federation_nodes + [self.me]:
                raise exceptions.OperationError('Only federation nodes can use the operation `CREATE`')

        else:
            # check if the input exists, is owned by the current_owner
            if not transaction['transaction']['input']:
                raise ValueError('Only `CREATE` transactions can have null inputs')

            tx_input = self.get_transaction(transaction['transaction']['input'])
            if not tx_input:
                raise exceptions.TransactionDoesNotExist('input `{}` does not exist in the bigchain'.format(
                    transaction['transaction']['input']))

            if tx_input['transaction']['new_owner'] != transaction['transaction']['current_owner']:
                raise exceptions.TransactionOwnerError('current_owner `{}` does not own the input `{}`'.format(
                    transaction['transaction']['current_owner'], transaction['transaction']['input']))

            # check if the input was already spent
            spent = self.get_spent(tx_input['id'])
            if spent:
                raise exceptions.DoubleSpend('input `{}` was already spent'.format(
                    transaction['transaction']['input']))

        # Check hash of the transaction
        calculated_hash = hash_data(self.serialize(transaction['transaction']))
        if calculated_hash != transaction['id']:
            raise exceptions.InvalidHash()

        # Check signature
        if not self.verify_signature(transaction):
            raise exceptions.InvalidSignature()

        return transaction

    def is_valid_transaction(self, transaction):
        """Check whether a transacion is valid or invalid.

        Similar to `validate_transaction` but does not raise an exception if the transaction is valid.

        Args:
            transaction (dict): transaction to check.

        Returns:
            bool: `True` if the transaction is valid, `False` otherwise

        """
        try:
            self.validate_transaction(transaction)
            return transaction
        except (ValueError, exceptions.OperationError, exceptions.TransactionDoesNotExist,
                exceptions.TransactionOwnerError, exceptions.DoubleSpend,
                exceptions.InvalidHash, exceptions.InvalidSignature):
            return False

    def create_block(self, validated_transactions):
        """Creates a block given a list of `validated_transactions`.

        Note that this method does not validate the transactions. Transactions should be validated before
        calling create_block.

        Args:
            validated_transactions (list): list of validated transactions.

        Returns:
            dict: created block.

        """
        # Create the new block
        block = {
            'timestamp': self.timestamp(),
            'transactions': validated_transactions,
            'node_pubkey': self.me,
            'voters': self.federation_nodes + [self.me]
        }

        # Calculate the hash of the new block
        block_data = self.serialize(block)
        block_hash = hash_data(block_data)
        block_signature = PrivateKey(self.me_private).sign(block_data)

        block = {
            'id': block_hash,
            'block': block,
            'signature': block_signature,
            'votes': []
        }

        return block

    # TODO: check that the votings structure is correctly constructed
    def validate_block(self, block):
        """Validate a block.

        Args:
            block (dict): block to validate.

        Returns:
            The block if the block is valid else it raises and exception
            describing the reason why the block is invalid.

        """

        # 1. Check if current hash is correct
        calculated_hash = hash_data(self.serialize(block['block']))
        if calculated_hash != block['id']:
            raise exceptions.InvalidHash()

        # 2. Validate all transactions in the block
        for transaction in block['block']['transactions']:
            if not self.is_valid_transaction(transaction):
                # this will raise the exception
                self.validate_transaction(transaction)

        return block

    def is_valid_block(self, block):
        """Check whether a block is valid or invalid.

        Similar to `validate_block` but does not raise an exception if the block is invalid.

        Args:
            block (dict): block to check.

        Returns:
            bool: `True` if the block is valid, `False` otherwise.

        """
        try:
            self.validate_block(block)
            return True
        except Exception:
            return False

    def write_block(self, block, durability='soft'):
        """Write a block to bigchain.

        Args:
            block (dict): block to write to bigchain.

        """
        block_serialized = rapidjson.dumps(block)
        r.table('bigchain').insert(r.json(block_serialized), durability=durability).run(self.conn)

    # TODO: Decide if we need this method
    def transaction_exists(self, transaction_id):
        response = r.table('bigchain').get_all(transaction_id, index='transaction_id').run(self.conn)
        return True if len(response.items) > 0 else False

    def create_genesis_block(self):
        """Create the genesis block

        Block created when bigchain is first initialized. This method is not atomic, there might be concurrency
        problems if multiple instances try to write the genesis block when the BigchainDB Federation is started,
        but it's a highly unlikely scenario.
        """

        # 1. create one transaction
        # 2. create the block with one transaction
        # 3. write the block to the bigchain


        blocks_count = r.table('bigchain').count().run(self.conn)

        if blocks_count:
            raise GenesisBlockAlreadyExistsError('Cannot create the Genesis block')


        payload = {'message': 'Hello World from the Bigchain'}
        transaction = self.create_transaction(self.me, self.me, None, 'GENESIS', payload=payload)
        transaction_signed = self.sign_transaction(transaction, self.me_private)

        # create the block
        block = self.create_block([transaction_signed])
        # add block number before writing
        block['block_number'] = 0
        self.write_block(block, durability='hard')

        return block

    def vote(self, block, previous_block_id, decision, invalid_reason=None):
        """Cast your vote on the block given the previous_block_hash and the decision (valid/invalid)
        return the block to the updated in the database.

        Args:
            block (dict): Block to vote.
            previous_block_id (str): The id of the previous block.
            decision (bool): Whether the block is valid or invalid.
            invalid_reason (Optional[str]): Reason the block is invalid

        """
        vote = {
            'voting_for_block': block['id'],
            'previous_block': previous_block_id,
            'is_block_valid': decision,
            'invalid_reason': invalid_reason,
            'timestamp': self.timestamp()
        }

        vote_data = self.serialize(vote)
        signature = PrivateKey(self.me_private).sign(vote_data)

        vote_signed = {
            'node_pubkey': self.me,
            'signature': signature,
            'vote': vote
        }

        return vote_signed

    def write_vote(self, block, vote, block_number):
        """
        Write the vote to the database
        """
        update = {'votes': r.row['votes'].append(vote)}

        # We need to *not* override the existing block_number, if any
        # FIXME: MIGHT HAVE RACE CONDITIONS WITH THE OTHER NODES IN THE FEDERATION
        if 'block_number' not in block:
            update['block_number'] = block_number

        r.table('bigchain')\
         .get(vote['vote']['voting_for_block'])\
         .update(update)\
         .run(self.conn)

    def get_last_voted_block(self):
        """
        Returns the last block that this node voted on
        """
        # query bigchain for all blocks this node is a voter but didn't voted on
        last_voted = r.table('bigchain')\
            .filter(r.row['block']['voters'].contains(self.me))\
            .filter(lambda doc: doc['votes'].contains(lambda vote: vote['node_pubkey'] == self.me))\
            .order_by(r.desc('block_number'))\
            .limit(1)\
            .run(self.conn)

        # return last vote if last vote exists else return Genesis block
        last_voted = list(last_voted)
        if not last_voted:
            return list(r.table('bigchain')
                         .filter(r.row['block_number'] == 0)
                         .run(self.conn))[0]

        return last_voted[0]

    def get_unvoted_blocks(self):
        """
        Return all the blocks that has not been voted by this node.
        """

        unvoted = r.table('bigchain')\
            .filter(lambda doc: doc['votes'].contains(lambda vote: vote['node_pubkey'] == self.me).not_())\
            .order_by(r.asc((r.row['block']['timestamp'])))\
            .run(self.conn)

        if unvoted and unvoted[0].get('block_number') == 0:
            unvoted.pop(0)

        return unvoted

    @staticmethod
    def serialize(data):
        """Static method used to serialize a dict into a JSON formatted string.

        This method enforces rules like the separator and order of keys. This ensures that all dicts
        are serialized in the same way.

        This is specially important for hashing data. We need to make sure that everyone serializes their data
        in the same way so that we do not have hash mismatches for the same structure due to serialization
        differences.

        Args:
            data (dict): dict to serialize

        Returns:
            str: JSON formatted string

        """
        return json.dumps(data, skipkeys=False, ensure_ascii=False,
                          separators=(',', ':'), sort_keys=True)

    @staticmethod
    def deserialize(data):
        """Static method used to deserialize a JSON formatted string into a dict.

        Args:
            data (str): JSON formatted string.

        Returns:
            dict: dict resulting from the serialization of a JSON formatted string.

        """
        return json.loads(data, encoding="utf-8")

    @staticmethod
    def timestamp():
        """Static method to calculate a UTC timestamp with microsecond precision.

        Returns:
            str: UTC timestamp.

        """
        dt = datetime.utcnow()
        return "{0:.6f}".format(time.mktime(dt.timetuple()) + dt.microsecond / 1e6)

    @staticmethod
    def generate_keys():
        """Generates a key pair.

        Returns:
            tuple: `(private_key, public_key)`. ECDSA key pair using the secp256k1 curve encoded
            in base58.

        """
        # generates and returns the keys serialized in hex
        return generate_key_pair()



# Separate all crypto code so that we can easily test several implementations

import hashlib
import sha3
import binascii
import base58
import bitcoin

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature


class PrivateKey(object):
    """
    PrivateKey instance
    """

    def __init__(self, key):
        """
        Instantiate the private key with the private_value encoded in base58
        """
        private_value = self.decode(key)
        private_numbers = self._private_value_to_cryptography_private_numbers(private_value)
        self.private_key = self._cryptography_private_key_from_private_numbers(private_numbers)

    def sign(self, data):
        """
        Sign data with private key
        """
        signer = self.private_key.signer(ec.ECDSA(hashes.SHA256()))
        signer.update(data.encode('utf-8'))
        signature = signer.finalize()
        return binascii.hexlify(signature).decode('utf-8')


    @staticmethod
    def encode(private_value):
        """
        Encode the decimal number private_value to base58
        """
        private_value_hex = bitcoin.encode_privkey(private_value, 'hex')
        private_value_base58 = base58.b58encode(bytes.fromhex(private_value_hex))
        return private_value_base58

    @staticmethod
    def decode(key):
        """
        Decode the base58 private_value to decimale
        """
        private_value_hex = binascii.hexlify(base58.b58decode(key))
        private_value = bitcoin.decode_privkey(private_value_hex)
        return private_value

    def _private_value_to_public_values(self, private_value):
        """
        Return the public values from the private value
        """
        public_value_x, public_value_y = bitcoin.privkey_to_pubkey(private_value)
        return (public_value_x, public_value_y)

    def _private_value_to_cryptography_private_numbers(self, private_value):
        """
        Return an instance of cryptography PrivateNumbers from the decimal private_value
        """
        public_value_x, public_value_y = self._private_value_to_public_values(private_value)
        public_numbers = PublicKey._public_values_to_cryptography_public_numbers(public_value_x, public_value_y)
        private_numbers = ec.EllipticCurvePrivateNumbers(private_value, public_numbers)
        return private_numbers

    @staticmethod
    def _cryptography_private_key_from_private_numbers(private_numbers):
        """
        Return an instace of cryptography PrivateKey from a cryptography instance of PrivateNumbers
        """
        return private_numbers.private_key(default_backend())


class PublicKey(object):

    def __init__(self, key):
        """
        Instantiate the public key with the compressed public value encoded in base58
        """
        public_value_x, public_value_y = self.decode(key)
        public_numbers = self._public_values_to_cryptography_public_numbers(public_value_x, public_value_y)
        self.public_key = self._criptography_public_key_from_public_numbers(public_numbers)

    def verify(self, data, signature):
        verifier = self.public_key.verifier(binascii.unhexlify(signature), ec.ECDSA(hashes.SHA256()))
        verifier.update(data.encode('utf-8'))
        try:
            verifier.verify()
        except InvalidSignature:
            return False

        return True

    @staticmethod
    def encode(public_value_x, public_value_y):
        """
        Encode the public key represented by the decimal values x and y to base58
        """
        public_value_compressed_hex = bitcoin.encode_pubkey([public_value_x, public_value_y], 'hex_compressed')
        public_value_compressed_base58 = base58.b58encode(bytes.fromhex(public_value_compressed_hex))
        return public_value_compressed_base58

    @staticmethod
    def decode(public_value_compressed_base58):
        """
        Decode the base58 public_value to the decimal x and y values
        """
        public_value_compressed_hex = binascii.hexlify(base58.b58decode(public_value_compressed_base58))
        public_value_x, public_value_y = bitcoin.decode_pubkey(public_value_compressed_hex.decode())
        return (public_value_x, public_value_y)

    @staticmethod
    def _public_values_to_cryptography_public_numbers(public_value_x, public_value_y):
        """
        Return an instance of cryptography PublicNumbers from the decimal x and y values
        """
        public_numbers = ec.EllipticCurvePublicNumbers(public_value_x, public_value_y, ec.SECP256K1())
        return public_numbers

    def _criptography_public_key_from_public_numbers(self, public_numbers):
        """
        Return an instance of cryptography PublicKey from a cryptography instance of PublicNumbers
        """
        return public_numbers.public_key(default_backend())


def generate_key_pair():
    """
    Generate a new key pair and return the pair encoded in base58
    """
    # Private key
    private_key = ec.generate_private_key(ec.SECP256K1, default_backend())
    private_value = private_key.private_numbers().private_value
    private_value_base58 = PrivateKey.encode(private_value)

    # Public key
    public_key = private_key.public_key()
    public_value_x, public_value_y = public_key.public_numbers().x, public_key.public_numbers().y
    public_value_compressed_base58 = PublicKey.encode(public_value_x, public_value_y)

    return (private_value_base58, public_value_compressed_base58)


def hash_data(data):
    return hashlib.sha3_256(data.encode()).hexdigest()
